#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>
#include <linux/swap.h>
#include <net/ip_masq.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/sysctl.h>
#include <linux/ip_fw.h>
#include <linux/ip_masq.h>
#include <linux/proc_fs.h>

#include <linux/inetdevice.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/route.h>
#include <net/ip_vs.h>
#include <net/ip_mycs.h>
#include <net/ip_vs_cb_pa.h>

#define   CR  '\r'
#define   LF  '\n'
#define   CRLF  "\r\n"
#define   CRLFCRLF  "\r\n\r\n"

__u32 ip_cs_norule(struct sk_buff *skb);
__u32 (*rule_matching)(struct sk_buff *skb);

EXPORT_SYMBOL(ip_cs_norule);
EXPORT_SYMBOL(rule_matching);

extern void ip_vs_cb_tcp_v4_send_reset(struct sk_buff *skb);
extern struct sock* ip_vs_cb_tcp_v4_lookup(struct tcphdr *th, struct sk_buff *skb);
extern void ip_vs_cb_tcp_measure_rcv_mss(struct sock *sk, struct sk_buff *skb);
char *mstrstr(char *source,  char *substr, size_t n);
int  extract_htp_cnt_len(struct ip_masq **ms_p);
extern void ip_vs_cb_tcp_reset(struct sock *sk);
static __inline__ int proto_doff(unsigned proto, char *th, unsigned size);
void  ip_vs_cb_adj_csum(struct sk_buff **skb_p, struct iphdr *iph, struct tcphdr *h);
struct ip_vs_dest *ip_vs_cb_get_dest(struct ip_masq *ms, struct ip_vs_service *svc);

#define  IP_MASQ_F_VS_CB_PARTIAL        0x0   /* Client request is not completely received*/
#define  IP_MASQ_F_VS_CB_COMPLETE       0x1   /* Client request is completely received*/

struct ip_masq *ip_masq_new_vs_cb(struct ip_vs_service *svc, struct sk_buff *skb)
{
  struct ip_masq *ms = NULL;
  struct ip_vs_cb_pa_service *pa_svc = NULL;
  struct iphdr	*iph = NULL;
  struct ip_vs_dest *dest = NULL;
  __u16 *portp = NULL;
   struct sk_buff *skb1 = NULL;

  if(!svc)
  {
    IP_VS_CB_MSG("ip_masq_new_vs_cb: NULL svc\n");
    return NULL;
  }

  if(!skb)
  {
    IP_VS_CB_MSG("ip_masq_new_vs_cb: NULL skb\n");
    return NULL;
  }

  iph = skb->nh.iph;
  if(!iph)
  {
    IP_VS_CB_MSG("ip_masq_new_vs_cb: NULL iphr\n");
    return NULL;
  }

  portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);
  if(!portp)
  {
    IP_VS_CB_MSG("ip_masq_new_vs_cb: NULL portp\n");
    return NULL;
  }

  read_lock(&__ip_vs_lock);

 /*
  *  Create a masquerading entry for content
  *  based virtual service
  */

  ms = ip_masq_new_vs(iph->protocol,
                      iph->daddr, portp[1],
                      0, 0,  /*saddr = 0, sport = 0 for time being as no real server is selected */
                      iph->saddr, portp[0], 0);
  if(ms != NULL)
  {
     ms->dest = NULL;
     ms->in_seq.init_seq  = 0;
     ms->in_seq.delta  = 0;
     ms->in_seq.previous_delta  = 0;

     ms->out_seq.init_seq  = 0;
     ms->out_seq.delta  = 0;
     ms->out_seq.previous_delta  = 0;
     ms->app_data = NULL;
     ms->timeout = svc->timeout;
     ms->last_sent_skb = NULL;
     /*
      * Set flag to denote that real server has not been 
      * chosen yet.
      */
     ms->cs_flags |= IP_MASQ_F_VS_CB_RSLOOSE;
     ms->tot_data_len = 0;
     ms->first_chunk = 1;
     memset(ms->http_mthd_nm, '\0', sizeof(ms->http_mthd_nm));
     ms->expected_data_len = 0;
     ms->htp_cnt_len = 0;
     memset(ms->trail_data, '\0', sizeof(ms->trail_data));
     skb_queue_head_init(&(ms->cli_req_data_head));

     ms->flags |= IP_VS_SVC_CONTENT_BASED;

     ms->first_resp = 1;
     memset(&(ms->start_tm), 0, sizeof(ms->start_tm));
     memset(&(ms->end_tm), 0, sizeof(ms->end_tm));

     ms->pa_svc = NULL;
     ms->cs_sk_clean = 0;
     /*
      * Look for a pre-exisiting entry
      */
/** uncomment normal code **/

     pa_svc = ip_vs_cb_pa_lookup_service(iph->protocol, iph->saddr, portp[1]);

     if(pa_svc)
     {
       if(pa_svc->rs_addr > 0)
       {
         dest = ip_vs_cb_dest_lookup(svc, ntohl(pa_svc->rs_addr));
         if(dest)
         {
           ms->cs_flags |= IP_MASQ_F_VS_CB_PRE_ALLOC;
           ms->saddr = pa_svc->rs_addr;
           ms->sport = pa_svc->rs_port;
           ip_vs_bind_masq(ms, dest);
           ms->pa_svc = pa_svc;
           atomic_inc(&(ms->pa_svc->refcnt));
         }
       }
     }

/**uncomment  normal code **/

/***code for pa all miss case 

  dest = ip_vs_cb_dest_lookup(svc, 0x80c6c0c6);
  ms->saddr = dest->addr;
  ms->sport = dest->port;
  ms->cs_flags |= IP_MASQ_F_VS_CB_PRE_ALLOC;
  ip_vs_bind_masq(ms, dest);
  ms->pa_svc = ip_vs_cb_pa_add_svc(ms->protocol, ms->daddr, ms->saddr, ms->sport);
  atomic_inc(&(ms->pa_svc->refcnt));

code for pa all miss case***/

     skb1 = skb_copy(skb, GFP_ATOMIC);

     if(skb1 == NULL)
     {
       IP_VS_CB_MSG("Schedule: skb_copy failed.\n");
       read_unlock(&__ip_vs_lock);
       return NULL;
     }

    /*
     * Store the initial SYN containing packet as
     * it will be sent to real server once data
     * is completely received
     */
     skb_queue_head(&(ms->cli_req_data_head), skb1);
  }
            
  read_unlock(&__ip_vs_lock);

  return ms;
}

int ip_vs_cb_queue_client_pkt(struct ip_masq *ms, struct sk_buff *skb)
{
   struct iphdr  *iph = skb->nh.iph;
   union ip_masq_tphdr h;
   int data_chunk_len = 0;
   int trail_data_len = 0;
   __u16 *portp = NULL;
   __u16  req_status;
   char *p  = NULL;
   char *data_p = NULL;
   struct sk_buff  *skb1 = NULL;

   req_status = IP_MASQ_F_VS_CB_PARTIAL;

   read_lock(&__ip_vs_lock);

   if(!iph)
   {
     IP_VS_CB_MSG("ip_vs_cb_queue_client_pkt(): NULL iph\n");
     return -2;
   }

   h.raw = ((char*) iph) + iph->ihl * 4;

   if(!(h.raw))
   {
     IP_VS_CB_MSG("ip_vs_cb_queue_client_pkt(): NULL h.raw\n");
     return -2;
   }

   portp = (__u16 *)h.raw;

   if(!portp)
   {
     IP_VS_CB_MSG("ip_vs_cb_queue_client_pkt(): NULL portp\n");
     return -2;
   }

  /*
   * Store initial CS generated sequence number from client ACK
   */
   if((h.th->ack) && (ms->in_seq.init_seq == 0))
   {
     ms->in_seq.init_seq  = ntohl(h.th->ack_seq) - 1;
   } 

   /*
    * Queue data until all client data is not received
    */
   skb1 = skb_copy(skb, GFP_ATOMIC);
   if(skb1 == NULL)
   {
     IP_VS_CB_MSG("ip_vs_cb_queue_client_pkt(1): Failed to copy skb\n");
     return -2;
   }
   else
     skb_queue_head(&(ms->cli_req_data_head), skb1);

   if(!(h.th->syn))
   {
     /*
      *  update data len from this skb chunk only for non-SYN
      *  packets as they conttain the actual data
      */
      __u16 tcp_hdr_len = h.th->doff*4;
      data_p =  h.raw + tcp_hdr_len;

      data_chunk_len = skb->len - (iph->ihl*4 + tcp_hdr_len);

      if(data_chunk_len > 0)
      {
        if(ms->first_chunk)
        {
          ms->tot_data_len += skb->len;
          /* 
           * Extract the HTTP request method from the first chunk 
           */ 
          p = memchr(data_p, ' ', data_chunk_len);

          if(p)
          {
            memcpy(ms->http_mthd_nm, data_p, p - data_p);
            ms->http_mthd_nm[p - data_p] = '\0';
          }
          ms->first_chunk = 0;
        }
        else
          ms->tot_data_len += data_chunk_len;

        if(!strcmp(ms->http_mthd_nm, "GET"))
        {
          /*
           * Look for CRLFCRLF sequence in the last 4 bytes
           */
           if(!memcmp(data_p + (data_chunk_len - 4), CRLFCRLF, 4))
              req_status = IP_MASQ_F_VS_CB_COMPLETE;
        }
        else
        if(!strcmp(ms->http_mthd_nm, "POST"))
        {
          if(ms->htp_cnt_len > 0)
          {
            /*
             * Http content length has been seen
             */
             ms->expected_data_len -= data_chunk_len;
             if(ms->expected_data_len <= 0)
             {
               req_status =  IP_MASQ_F_VS_CB_COMPLETE;
               goto req_complete;
             }
          }
          else
          if((trail_data_len = strlen(ms->trail_data)) > 0)
          {
            if(!memcmp(data_p, ms->trail_data, trail_data_len))
            {
              /*
               * A match for CRLFCRLF sequence whose parts spanned 
               * across multiple segments has been found
               */
               if(extract_htp_cnt_len(&ms) < 0)
               {
                 IP_VS_CB_MSG("ip_vs_cb_queue_client_pkt(): extract_htp_cnt_len() returned error\n"); 
                 read_unlock(&__ip_vs_lock);
                 return  -2;
               }
     
               if(ms->expected_data_len <= 0)
               {
                 req_status =  IP_MASQ_F_VS_CB_COMPLETE;
                 goto req_complete;
               }
            }
          }
 
          if((p = mstrstr(data_p, CRLFCRLF, data_chunk_len)) != NULL)
          {
            if(extract_htp_cnt_len(&ms) < 0)
            {
              IP_VS_CB_MSG("ip_vs_cb_queue_client_pkt(): extract_htp_cnt_len() returned error\n"); 
               read_unlock(&__ip_vs_lock);
               return  -2;
            }

            if(ms->expected_data_len <= 0)
            {
               req_status =  IP_MASQ_F_VS_CB_COMPLETE;
               goto req_complete;
            }
          }
          else
          {
            memcpy(ms->trail_data, data_p + (data_chunk_len - 4), 4);
 
            if((p = (char *) memchr(ms->trail_data, CR, 4)) != NULL)
            {
              /*
               * If some trace of CRLFCRLF is present in this segment
               */
              ms->trail_data[0] = '\0';
 
              switch(p - ms->trail_data)
              {
                /*
                 *  In all following cases below  we are storing the 
                 *  the anticipated remaining bytes of the CRLFCRLF 
                 *  sequence in ms->trail_data, if ms->trail_data[0] == '\0' 
                 *  in consequence to following checks, it means
                 *  part of CRLFCRLF sequence is not lying in this
                 *  segment as we suspected by being in this if block
                 */
                case 1:
        
                 if((ms->trail_data[2] == LF) && (ms->trail_data[3] == CR))
                 {
                   ms->trail_data[0] = LF;
                   ms->trail_data[1] = '\0';
                 }
                 break;

                case 2:
                 if(ms->trail_data[3] == LF)
                   strcpy(ms->trail_data, CRLF);
                 break;

                case 3:
                 ms->trail_data[0] = LF;
                 ms->trail_data[1] = CR;
                 ms->trail_data[2] = LF;
                 ms->trail_data[3] = '\0';
              }
            }
            else
              ms->trail_data[0] = '\0';
      
            ms->trail_data[4] = '\0';
          }
        } /* if(!strcmp( */
        else
        {
          req_status = IP_MASQ_F_VS_CB_COMPLETE;
        }
      }
   }
   /*
    * if all data not received 
    *  return 0
    * else
    *  return  1
   */
   if(req_status != IP_MASQ_F_VS_CB_COMPLETE)
   {
     if((data_chunk_len == 0) && (h.th->ack) && (ms->first_chunk))
     {
       read_unlock(&__ip_vs_lock);
       return 2;
     }
       
     read_unlock(&__ip_vs_lock);
     return 0;
   }


req_complete:

   read_unlock(&__ip_vs_lock);

   if(ms->cs_flags & IP_MASQ_F_VS_CB_PRE_ALLOC) 
   {
      ms->cs_flags &= ~IP_MASQ_F_VS_CB_RSLOOSE;
      ms->cs_flags |= IP_MASQ_F_VS_CB_RSNOTREADY;
   } 

   return 1;
}
  
int ip_vs_cb_schedule(struct ip_masq *ms, struct ip_vs_service *svc)
{
   struct ip_vs_dest *dest = NULL;
   /*
    *  This function assumes that the client request 
    *  has been completely received
    */
   read_lock(&__ip_vs_lock);

   dest = ip_vs_cb_get_dest(ms, svc);

   if(dest == NULL)
   {
     /*
      * No luck at this point
      */
     skb_queue_purge(&(ms->cli_req_data_head));
     read_unlock(&__ip_vs_lock);
     return -1;
   }

   ip_vs_cb_bind_dest(ms, dest, 0);

   read_unlock(&__ip_vs_lock);
   
   return 0;
}

int ip_vs_cb_send_client_ack(struct ip_masq *ms, struct sk_buff  *skb)
{
  struct tcphdr *th = NULL;
  struct sock *sk = NULL;
  struct tcp_opt *tp = NULL;
  unsigned short len = 0;

  skb->h.raw = skb->nh.raw + (skb->nh.iph)->ihl*4;

  th = skb->h.th;

  __skb_pull(skb, skb->h.raw - skb->data);
  len = skb->len;

  TCP_SKB_CB(skb)->seq = ntohl(th->seq);
  TCP_SKB_CB(skb)->end_seq = (TCP_SKB_CB(skb)->seq + th->syn + th->fin +
                              len - th->doff*4);
  TCP_SKB_CB(skb)->ack_seq = ntohl(th->ack_seq);

  sk = ip_vs_cb_tcp_v4_lookup(th, skb);
  if(!sk)
  {
    IP_VS_CB_MSG("MS=%u ip_vs_cb_send_client_ack: No sock structure found\n", (unsigned int)ms);
    return -1;
  }

  tp = &(sk->tp_pinfo.af_tcp);

  __skb_pull(skb, th->doff*4);

  ip_vs_cb_tcp_measure_rcv_mss(sk, skb);

  tp->rcv_nxt = TCP_SKB_CB(skb)->end_seq;

  tcp_send_ack(sk);

  if(skb)
    kfree_skb(skb);
  return 0;
}

void ip_vs_cb_dequeue_client_pkt(struct ip_masq *ms, struct sk_buff  **skb_p)
{
  struct sk_buff *skb = NULL;
  struct iphdr  *svr_iph = NULL;
  union ip_masq_tphdr svr_tp;
  struct iphdr  *cli_iph = NULL;
  union ip_masq_tphdr cli_tp;

  read_lock(&__ip_vs_lock);

  if((ms->cs_flags & IP_MASQ_F_VS_CB_CLI_DATA_FWD) && ms->last_sent_skb)
  {
    svr_iph = (*skb_p)->nh.iph;
    svr_tp.raw = (char*) svr_iph + svr_iph->ihl * 4;

    cli_iph = ms->last_sent_skb->nh.iph;
    cli_tp.raw = (char*) cli_iph + cli_iph->ihl * 4;

    if(ntohl(svr_tp.th->ack_seq) <= ntohl(cli_tp.th->seq)) 
    {
      skb = skb_copy(ms->last_sent_skb, GFP_ATOMIC);
    }
    else
    {
      if(ms->last_sent_skb)
        kfree_skb(ms->last_sent_skb);

      skb = skb_dequeue_tail(&(ms->cli_req_data_head));
      ms->last_sent_skb = skb_copy(skb, GFP_ATOMIC);
    }
  }
  else
  {
    if(ms->last_sent_skb)
      kfree_skb(ms->last_sent_skb);

    skb = skb_dequeue_tail(&(ms->cli_req_data_head));
    ms->last_sent_skb = skb_copy(skb, GFP_ATOMIC);
  }

  if(skb != NULL)
  {
    if(!(*skb_p)) 
    {
      kfree_skb(*skb_p);
      *skb_p = NULL;
    }
    *skb_p = skb;
  }
  read_unlock(&__ip_vs_lock);
}

void ip_vs_cb_send_client_rst(struct sk_buff  *skb)
{
  struct tcphdr *th;
  unsigned short len;

  skb->h.raw = skb->nh.raw + skb->nh.iph->ihl*4;

  th = skb->h.th;
  len = skb->len;
  __skb_pull(skb, skb->h.raw - skb->data);

  TCP_SKB_CB(skb)->seq = ntohl(th->seq);
  TCP_SKB_CB(skb)->end_seq = (TCP_SKB_CB(skb)->seq + th->syn + th->fin + len - th->doff*4);
  TCP_SKB_CB(skb)->ack_seq = ntohl(th->ack_seq);

  skb->used = 0;

  ip_vs_cb_tcp_v4_send_reset(skb);
}


void ip_vs_cb_fix_cli_ack_seq (struct ip_masq  *ms, struct sk_buff **skb_p)
{
  struct iphdr  *iph = (*skb_p)->nh.iph;
  union ip_masq_tphdr h;
  __u32 temp_seq = 0;

  h.raw = (char*) iph + iph->ihl * 4;

  temp_seq = ntohl(h.th->ack_seq) - ms->in_seq.init_seq + ms->out_seq.init_seq;

  h.th->ack_seq = htonl(temp_seq);
}

void ip_vs_cb_get_server_seq(struct ip_masq  *ms, struct sk_buff *skb)
{
   struct iphdr  *iph = skb->nh.iph;
   union ip_masq_tphdr h;

   read_lock(&__ip_vs_lock);

   h.raw = (char*) iph + iph->ihl * 4;

  /*
   * Store initial Server generated sequence number from Server SYN/ACK
   */
   if((h.th->ack) && (ms->out_seq.init_seq == 0))
   {
     ms->out_seq.init_seq  = ntohl(h.th->seq);
     ms->cs_flags |= IP_MASQ_F_VS_CB_CLI_DATA_FWD;
   } 

   read_unlock(&__ip_vs_lock);
} 

void ip_vs_cb_fix_svr_dat_seq(struct ip_masq *ms, struct sk_buff **skb_p)
{
  struct iphdr  *iph = (*skb_p)->nh.iph;
  union ip_masq_tphdr h;
  __u32 temp_seq = 0;

  read_lock(&__ip_vs_lock);

  h.raw = (char*) iph + iph->ihl * 4;

  temp_seq = ntohl(h.th->seq) - ms->out_seq.init_seq + ms->in_seq.init_seq;

  h.th->seq = htonl(temp_seq);

  read_unlock(&__ip_vs_lock);
}


struct sk_buff*  skb_concat(struct sk_buff_head *h, int newskb_len, __u16 whence)
{
  /*
   * This function returns a single sk_buff, having 'data' (excluding ip/tcp hdrs)
   * merged into a single buffer, the merging begins from tail end
   * to the head. The number of sk_buff nodes specified in 'whence' are skipped
   * relative to the tail
   */
   struct sk_buff *prv = h->prev;
   struct iphdr  *iph = NULL;
   union ip_masq_tphdr tph;
   int   data_len = 0;
   __u16 skip_count = 0;
   int   mrg_data_len = 0;
   char  *data_p = NULL;
   char  *tmp_buf = NULL;
   struct sk_buff *newskb = NULL;
   __u16 first_buf = 1;

   newskb = alloc_skb(newskb_len, GFP_ATOMIC);

   if(!newskb)
   {
     IP_VS_CB_MSG("skb_concat(): Failed im skb_alloc, len requested = %d\n", newskb_len);     
     return NULL;
   }
   
   while(prv != NULL)
   {
     if(skip_count < whence)
     {
       prv = prv->prev;
       skip_count++;
       continue;
     }
      
     if(first_buf)
     {
       first_buf = 0;
       data_p = (char *) prv->nh.iph;
       data_len = prv->len;
     }
     else
     {
       iph = prv->nh.iph;
       tph.raw = (char*) iph + iph->ihl * 4;
       data_p = tph.raw + tph.th->doff * 4;
     
       data_len = prv->len - (data_p - (char*)prv->data);
     }
     
     if(data_len > 0)
     {
       if(data_len > skb_tailroom(newskb))
       {
         IP_VS_CB_MSG("skb_concat(): node len = %d whereas tailroom = %d, data merged = %d\n", 
                   data_len, skb_tailroom(newskb), mrg_data_len);     
         kfree_skb(newskb);
         return NULL;
       }
       tmp_buf = skb_put(newskb, data_len);
       memcpy(tmp_buf, data_p, data_len);
       mrg_data_len += data_len;

     }
     
     prv = prv->prev;
   
     if(prv == h->next->prev)
     /*
      * We have looped through the doubly linked list
      */
       break;
   }

   if(newskb)
   {
     newskb->nh.raw = newskb->data;
   }

   return newskb;
}

struct ip_vs_dest *ip_vs_cb_dest_lookup(struct ip_vs_service *service, __u32 addr)
{
  register struct list_head *p, *q;
  struct ip_vs_dest *dest=NULL;

  p = (struct list_head *)service->sched_data;
  p = p->next;
  q = p;

  do 
  {
    if (q == &service->destinations) 
    {
      q = q->next;
      continue;
    }
    dest = list_entry(q, struct ip_vs_dest, n_list);
    if (ntohl(dest->addr)==addr)
      goto out;
    q = q->next;
  } while (q != p);

  dest = NULL;

  out:

  service->sched_data = q;

  return dest;
}

char *mstrstr(char *source,  char *substr, size_t n)
{
  char c = source[n - 1]; 
  char *p = NULL;
  int  subStrLen = 0;
  
  /*
   * Following is needed to make strstr work
   */
  source[n - 1] = '\0';

  p = strstr(source, substr);

  source[n - 1] = c;

  if(p == NULL)
  {
    subStrLen =  strlen(substr);
   
    if(!memcmp(source + n - subStrLen, substr, subStrLen))
      p = source + n - subStrLen;  
  }

  return p;
}

int  extract_htp_cnt_len(struct ip_masq **ms_p)
{
  struct sk_buff  *skb_mrg = NULL;
  char   *p = NULL;
  char   htp_cnt_len_str[20];
  int    i = 0;
  struct ip_masq *ms = NULL;
  __u16 tcp_hdr_len = 0;
  int    data_chunk_len = 0;
  char   *data_p =  NULL;
  struct iphdr  *iph = NULL;
  union ip_masq_tphdr h;

  ms = *ms_p;

  /*
   * Concatenate all packets received thus far
   * and determine content lengh
   */
  if(ms->cli_req_data_head.qlen == 3)
    skb_mrg = skb_peek(&(ms->cli_req_data_head));
  else    
    skb_mrg = skb_concat(&(ms->cli_req_data_head), ms->tot_data_len, 2);

  if(skb_mrg == NULL)
  {
     /*
      * It's a severe, error we have no alternative except to
      * drop the queued data
      */
     IP_VS_CB_MSG("extract_htp_cnt_len(): Failed to merge skbs\n");
     skb_queue_purge(&(ms->cli_req_data_head));
     return -1;
  }

  iph = skb_mrg->nh.iph;
  h.raw = ((char*) iph) + iph->ihl * 4;
  tcp_hdr_len = h.th->doff * 4;
  data_p =  h.raw + tcp_hdr_len;
  data_chunk_len = skb_mrg->len - (iph->ihl*4 + tcp_hdr_len);

  if(((p = mstrstr(data_p, "Content-length:", data_chunk_len)) != NULL) || 
     ((p = mstrstr(data_p, "Content-Length:", data_chunk_len)) != NULL))
  {
     p += 15;  /* place p past the ':' */

     while((*p) == ' ')
       p++;
     for(i = 0; (*p) != CR; p++, i++)
       htp_cnt_len_str[i] = *p;

     htp_cnt_len_str[i] = '\0';
    
     ms->htp_cnt_len = simple_strtol(&(htp_cnt_len_str[0]), (char **) NULL, 10);     

     if(ms->htp_cnt_len <= 0)
     {
       if(ms->cli_req_data_head.qlen != 3)
         kfree_skb(skb_mrg);
       IP_VS_CB_MSG("extract_htp_cnt_len(): HTTP content length = %d is invalid, htp_cnt_len_str = %s\n", ms->htp_cnt_len, htp_cnt_len_str);
       return -1;
     }
    
     /*
      * Expected data len is http content lenghth minus the data seen in this
      * chunk
      */
     ms->expected_data_len = (ms->htp_cnt_len) - ((data_p + data_chunk_len) - (p + 4));
  }
  else
  {
    if(ms->cli_req_data_head.qlen != 3)
      kfree_skb(skb_mrg);
    IP_VS_CB_MSG("extract_htp_cnt_len(): Fatal error Conten-Length tag must be present\n");
    return -1;
  }

  if(ms->cli_req_data_head.qlen != 3)
    kfree_skb(skb_mrg);
 
  return 0;
}

int ip_vs_cb_peek_serv_resp(struct ip_masq *ms, struct sk_buff *skb, struct ip_vs_dest **dest)
{
  struct iphdr  *iph = NULL;
  union ip_masq_tphdr h;
  int   data_len = 0;
  int   ret = 0;
  char  *p  = NULL;
  char  *p1  = NULL;
  char  *data_p = NULL;
  char  svr_resp_cd_str[20];
  __u16 tcp_hdr_len = 0;

  read_lock(&__ip_vs_lock);

  if(!skb)
  {
    IP_VS_CB_MSG("ip_vs_cb_peek_serv_resp(): skb==NULL\n");
    return -1;
  }

  iph = skb->nh.iph;

  if(!iph)
  {
    IP_VS_CB_MSG("ip_vs_cb_peek_serv_resp(): iph==NULL\n");
    return -1;
  }

  h.raw = ((char*) iph) + iph->ihl * 4;

  if(!(h.raw))
  {
    IP_VS_CB_MSG("ip_vs_cb_peek_serv_resp(): (h.raw)==NULL\n");
    return -1;
  }

  tcp_hdr_len = h.th->doff*4;
  data_p =  h.raw + tcp_hdr_len;

  if(!data_p)
  {
    IP_VS_CB_MSG("ip_vs_cb_peek_serv_resp(): data_p==NULL\n");
    return -1;
  }

  data_len = skb->len - (iph->ihl*4 + tcp_hdr_len);
   
  if(!(h.th->syn) && (data_len > 0))
  {
    /*
     * HTTP response looks like
     * HTTP/1.1 200 OK\r\n
     */
    if((p = memchr(data_p, ' ', data_len)) !=  NULL)
    {
      while(p && ((*p) == ' ')) p++;
      if(p && ((p1 = memchr(p, ' ', data_len - (p - data_p))) !=  NULL))
      {
        memcpy(svr_resp_cd_str, p, p1 - p);
        svr_resp_cd_str[p1 - p] = '\0';
      }
      else
        svr_resp_cd_str[0] = '\0';

      IP_VS_CB_PROF("resp_cd_str=%s ", svr_resp_cd_str);
     
      if(strcmp(svr_resp_cd_str, "200"))
      {
        /*
         * Preallocated real server gave a -ve response
         * Let's see what rule matching decision looks like
         */

         *dest = ip_vs_cb_get_dest(ms, NULL);
         if((*dest) == NULL)
         {
           IP_VS_CB_MSG("ip_vs_cb_peek_serv_resp(): ip_vs_cb_get_dest() returned error\n");
           goto out1;
         }

         
         if(iph->saddr == (*dest)->addr)
         {
           /*
            *  Rule matching decision is same as
            *  preallocate decision
            */
           goto out1;
         }
         else
         {
           ret = 1;
           IP_VS_CB_MSG2("old_rs==%x new_rs=%x\n", ntohl(iph->saddr), ntohl((*dest)->addr));
           goto out2;
         }
      }
      else
      {
        IP_VS_CB_MSG2("dest==%x\n", (unsigned int) (*dest));
       /*
        * Preallocate guess was right!
        */
        goto out1;
      }
    }
  }
  
/*
  if(data_len == 0)
    ret = -1;
*/

  goto out2;

out1:
  /*
   * Free up the client queued data as server sent OK
   */
  if(ms->cli_req_data_head.qlen > 0)
     skb_queue_purge(&(ms->cli_req_data_head));

  /*
   * Signal that real server is ready
   */
  ms->cs_flags &= ~IP_MASQ_F_VS_CB_RSNOTREADY;
  ms->cs_flags |= IP_MASQ_F_VS_CB_RSREADY;

out2:
  read_unlock(&__ip_vs_lock);

  return ret;
}


__u32 ip_cs_norule(struct sk_buff *skb)
{
  return 0; 
}

__initfunc(int ip_vs_cb_init(void))
{
  IP_VS_CB_MSG("Inserting Content Switch module in preallocate scheme\n");

  rule_matching = ip_cs_norule;   

  ip_vs_cb_pa_svc_init();

  return 0;
}

int ip_vs_cb_clean_cs_sock(struct ip_masq *ms, struct sk_buff  *skb)
{
  struct tcphdr *th = NULL;
  struct sock *sk = NULL;
  unsigned short len = 0;
  struct sk_buff  *skb_tmp = NULL;

  read_lock(&__ip_vs_lock);
  IP_VS_CB_MSG2("MS=%u Entering ip_vs_cb_clean_cs_sock ms->cs_sk_clean=%d\n", (unsigned int)ms, ms->cs_sk_clean);

  if(!(ms->cs_sk_clean))
  {
    skb_tmp = skb_copy(skb, GFP_ATOMIC);

    if(skb_tmp == NULL)
    {
      IP_VS_CB_MSG("MS=%u ip_vs_cb_clean_cs_sock: Failed to copy skb_tmp\n", (unsigned int)ms);
      read_unlock(&__ip_vs_lock);
      return -1;
    }

    skb_tmp->h.raw = skb_tmp->nh.raw + (skb_tmp->nh.iph)->ihl*4;

    th = skb_tmp->h.th;

    __skb_pull(skb_tmp, skb_tmp->h.raw - skb_tmp->data);
    len = skb_tmp->len;

    TCP_SKB_CB(skb_tmp)->seq = ntohl(th->seq);
    TCP_SKB_CB(skb_tmp)->end_seq = (TCP_SKB_CB(skb_tmp)->seq + th->syn + th->fin +
                              len - th->doff*4);
    TCP_SKB_CB(skb_tmp)->ack_seq = ntohl(th->ack_seq);

    sk = ip_vs_cb_tcp_v4_lookup(th, skb_tmp);
    if(!sk)
    {
      IP_VS_CB_MSG("MS=%u ip_vs_cb_send_client_ack: No sock structure found\n", (unsigned int)ms);
      return -1;
    }

    IP_VS_CB_MSG2("MS=%u SK=%u sock state =%d sk->prev=%u sk->timer.prev=%u expires=%u sk->rmem_alloc =%d sk->wmem_alloc=%d\n",
                 (unsigned int)ms,
                 (unsigned int)sk,
                 sk->state,
                 (unsigned int)sk->prev,
                 (unsigned int)(sk->timer.prev),
                 (unsigned int)(sk->timer.expires),
                 atomic_read(&sk->rmem_alloc),
                 atomic_read(&sk->wmem_alloc));

    ip_vs_cb_tcp_reset(sk);

    if(skb_tmp)
      kfree_skb(skb_tmp);

    ms->cs_sk_clean = 1;
  }

  IP_VS_CB_MSG2("MS=%u Leaving ip_vs_cb_clean_cs_sock\n", (unsigned int)ms );
  read_unlock(&__ip_vs_lock);

  return 0;
}

struct ip_vs_dest *ip_vs_cb_get_dest(struct ip_masq *ms, struct ip_vs_service *svc)
{
  struct ip_vs_dest *dest = NULL;
  struct sk_buff    *skb1 = NULL;
  __u32             rs_addr = 0;

  if(!svc)
  {
    /*
     * No problem!
     * Look for service now
     */
    svc = ip_vs_lookup_service(0, ms->protocol, ms->maddr, ms->mport);

    if(!svc)
    {
      IP_VS_CB_MSG("ip_vs_cb_get_dest() : No service found for masq proto=%d masq addr=%x masq port=%d", ms->protocol, ntohl(ms->maddr), ntohs(ms->mport));
      return NULL;
    }
  }       

  if(ms->cli_req_data_head.qlen)
  {
    /*
     * We are skipping first two skbs as they have SYN and ACK from
     * client and no data
     */
    skb1 = skb_concat(&(ms->cli_req_data_head), ms->tot_data_len, 2);

    if(skb1 == NULL)
    {
      /*
       * It's a severe error we have no alternative except to
       * drop the queued data
       */
      IP_VS_CB_MSG("ip_vs_cb_get_dest(): Failed to merge skbs\n");
      return NULL;
    }

    /*
     * Get the real server address from the rule matching
     * module
     */
    rs_addr = rule_matching(skb1);

    /*
     * We don't need this skb1 any more
     */
    if(skb1)
      kfree_skb(skb1);

    if(rs_addr == 0xffffffff)
    {
      /*
       * If the client address needs to be rejected do so
       */
      skb_queue_purge(&(ms->cli_req_data_head));
      return NULL;
    }

    if(rs_addr > 0)
      dest = ip_vs_cb_dest_lookup(svc, rs_addr);
  }

  if((rs_addr == 0) || (dest == NULL))
    /*
     * Find real server via default ip vs scheduler
     */
    dest = svc->scheduler->schedule(svc);

  return dest;
}

void ip_vs_cb_bind_dest(struct ip_masq *ms, struct ip_vs_dest *dest, int lock_flg)
{
  if(lock_flg)
    read_lock(&__ip_vs_lock);
  
  ms->saddr = dest->addr;
  ms->sport = dest->port;

  /* Un bind the old dest */
  if(!(ms->dest))
    ip_vs_unbind_masq(ms);
      
  ip_vs_bind_masq(ms, dest);

  atomic_inc(&(ms->dest->inactconns));
  ms->cs_flags &= ~IP_MASQ_F_VS_CB_RSLOOSE;
   /*
    * Set indication that implies real server is chosen
    * but not yet ready to accept data
    */
  ms->cs_flags |= IP_MASQ_F_VS_CB_RSNOTREADY;

  if(ms->pa_svc)
  {
     /*                                            
      * delete the entry of prior
      * real server from the hash table            
      */                                           
    ip_vs_cb_pa_del_svc(ms->pa_svc);              
  }

  /*
   * Store information of selected real server in pre-allocated
   * hash table
   */
  ms->pa_svc = ip_vs_cb_pa_add_svc(ms->protocol, ms->daddr, ms->saddr, ms->sport);

  if(lock_flg)
    read_unlock(&__ip_vs_lock);
}


struct sk_buff *ip_vs_cb_get_rst_for_svr(struct ip_masq *ms, struct sk_buff *skb)
{
  struct sk_buff   *rst_skb = NULL;
  struct iphdr  *svr_iph = skb->nh.iph;
  union ip_masq_tphdr svr_h;
  struct iphdr  *rst_iph = NULL;
  union ip_masq_tphdr rst_h;
  
  read_lock(&__ip_vs_lock);

  IP_VS_CB_MSG2("Entering ip_vs_cb_get_rst_for_svr\n");

  if((ms->cli_req_data_head.prev) != NULL)
  {
    rst_skb = skb_copy(ms->cli_req_data_head.prev, GFP_ATOMIC);

    if(rst_skb == NULL)
    {
      IP_VS_CB_MSG("ip_vs_cb_get_rst_for_svr: skb_copy for rst_skb failed.\n");
      read_unlock(&__ip_vs_lock);
      return NULL;
    }

    rst_iph  = rst_skb->nh.iph;

    rst_h.raw = ((char*) rst_iph) + rst_iph->ihl * 4;

    svr_h.raw = ((char*) svr_iph) + svr_iph->ihl * 4;
    
    if(svr_h.th->ack_seq)
    {
      rst_h.th->seq = svr_h.th->ack_seq;
      rst_h.th->ack_seq = 0;
    }
    else
    {
      rst_h.th->seq = 1;
      rst_h.th->ack_seq = svr_h.th->seq;
    }
     
    rst_h.th->syn = 0;
    rst_h.th->fin = 0;
    rst_h.th->rst = 1;

    IP_VS_CB_MSG2("ip_vs_cb_get_rst_for_svr(): Sending RST saddr=%x daddr=%x dest=%x rs_addr=%x sport=%d dport=%d seq=%u ack_seq=%u SYN=%d ACK=%d RST=%d FIN=%d skb->len=%d ip_len=%d\n",
        ntohl(rst_iph->saddr),
        ntohl(rst_iph->daddr),
        (unsigned int)(ms->dest),
        ntohl(ms->saddr),
        ntohs(rst_h.th->source),
        ntohs(rst_h.th->dest),
        ntohl(rst_h.th->seq),
        ntohl(rst_h.th->ack_seq),
        ((rst_h.th->syn) ? 1 : 0),
        ((rst_h.th->ack) ? 1 : 0),
        ((rst_h.th->rst) ? 1 : 0),
        ((rst_h.th->fin) ? 1 : 0),
        rst_skb->len,
        ntohs(rst_iph->tot_len));

    rst_skb->ip_summed = CHECKSUM_NONE;

    ip_vs_cb_adj_csum(&rst_skb, rst_iph, rst_h.th);

  }

  IP_VS_CB_MSG2("Leave ip_vs_cb_get_rst_for_svr rst_skb = %u\n", (unsigned int)rst_skb);

  read_unlock(&__ip_vs_lock);
 
  return   rst_skb;
}

static __inline__ int proto_doff(unsigned proto, char *th, unsigned size)
{
        int ret = -1;
        switch (proto) {
                case IPPROTO_ICMP:
                        if (size >= sizeof(struct icmphdr))
                                ret = sizeof(struct icmphdr);
                        break;
                case IPPROTO_UDP:
                        if (size >= sizeof(struct udphdr))
                                ret = sizeof(struct udphdr);
                        break;
                case IPPROTO_TCP:
                        /*
                        *       Is this case, this check _also_ avoids
                        *       touching an invalid pointer if
                        *       size is invalid
                        */
                        if (size >= sizeof(struct tcphdr)) {
                                ret = ((struct tcphdr*)th)->doff << 2;
                                if (ret > size) {
                                        ret = -1 ;
                                }
                        }

                        break;
                default:
                        /*      Other proto: nothing to say, by now :) */
                        ret = 0;
        }
        if (ret < 0)
                IP_VS_CB_MSG("mess proto_doff for proto=%d, size =%d\n",
                        proto, size);
        return ret;
}

void  ip_vs_cb_adj_csum(struct sk_buff **skb_p, struct iphdr *iph, struct tcphdr *h)
{
  struct sk_buff  *skb = *skb_p;
  unsigned short size;
  int doff = 0;
  int csum = 0;

  IP_VS_CB_MSG2("Entering ip_vs_cb_adj_csum()\n");
  /*
   *    IP payload size
   */
  size = ntohs(iph->tot_len) - (iph->ihl * 4);

  doff = proto_doff(iph->protocol, (unsigned char*)h, size);

  IP_VS_CB_MSG2("ip_vs_cb_adj_csum() size=%d doff=%d\n", size, doff);

  csum = csum_partial((unsigned char*)h + doff, size - doff, 0);
  skb->csum = csum;

  /*
   *Protocol csum
   */
  switch (iph->protocol)
  {
    case IPPROTO_TCP:
       h->check = 0;
       h->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
                                    size, iph->protocol,
                                    csum_partial((unsigned char*)h, doff, csum));
       break;

    case IPPROTO_UDP:
       h->check = 0;
       h->check = csum_tcpudp_magic(iph->saddr, iph->daddr,
                                    size, iph->protocol,
                                    csum_partial((unsigned char*)h, doff, csum));
       if(h->check == 0)
       h->check = 0xFFFF;
       break;
  }

  /*
   * Adjust IP Checksum
   */
  iph->check = 0;
  iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);

  IP_VS_CB_MSG2("Leaving ip_vs_cb_adj_csum()\n");
}
