/*
 * IP_VS_CB     An implementation of the IP virtual service content based 
 *              support for the LINUX operating system.  IPVSCB is 
 *              implemented as a part of IP masquerading code. 
 *              IPVSCB can be used to build highly available server based on a
 *              cluster of servers.
 *
 * Version:     
 *
 * Authors:     Chandra Prakash <cprakash@cs.uccs.edu>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 */
/**
#include <linux/config.h>    
#include <linux/module.h>    
#include <linux/types.h>     
#include <linux/kernel.h>    
#include <linux/errno.h>     
#include <linux/vmalloc.h>   
#include <linux/swap.h>      
#include <net/ip_masq.h>     
#ifdef CONFIG_KMOD           
#include <linux/kmod.h>      
#endif                       
#include <linux/sysctl.h>    
#include <linux/ip_masq.h>   
                             
#include <linux/ip.h>        
#include <net/ip.h>          
#include <net/tcp.h>         
#include <net/ip_vs.h>  
#include <net/ip_mycs.h>
#include <net/ip_vs_cb_pa.h>
**/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/vmalloc.h>
#include <linux/swap.h>
#include <net/ip_masq.h>
#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif
#include <linux/sysctl.h>
#include <linux/ip_fw.h>
#include <linux/ip_masq.h>
#include <linux/proc_fs.h>

#include <linux/inetdevice.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/route.h>
#include <net/ip_vs.h>
#include <net/ip_vs_cb_pa.h>
#include <net/ip_mycs.h>


#define IP_VS_CB_PA_TAB_BITS 8
#define IP_VS_CB_PA_TAB_SIZE (1 << IP_VS_CB_PA_TAB_BITS)
#define IP_VS_CB_PA_TAB_MASK (IP_VS_CB_PA_TAB_SIZE - 1)

struct list_head ip_vs_cb_pa_table[IP_VS_CB_PA_TAB_SIZE];

/*
 *      Returns hash value for virtual service
 */
static __inline__ unsigned
ip_vs_cb_pa_hashkey(unsigned proto, __u32 addr, __u16 port)
{
   register unsigned porth = ntohs(port);

   return (proto^ntohl(addr)^(porth>>IP_VS_CB_PA_TAB_BITS)^porth)
                & IP_VS_CB_PA_TAB_MASK;

}

/*
 *  Lookup service by {proto,addr,port} in the pre-allocate service table.
 */
struct ip_vs_cb_pa_service *
ip_vs_cb_pa_lookup_service(__u16 protocol, __u32 addr, __u16 port)
{
  unsigned hash;
  struct ip_vs_cb_pa_service *svc;
  struct list_head *l,*e;

  /*
   * Check for "full" addressed entries
   */
  hash = ip_vs_cb_pa_hashkey(protocol, addr, port);

  l = &ip_vs_cb_pa_table[hash];
  for (e=l->next; e!=l; e=e->next) 
  {
    svc = list_entry(e, struct ip_vs_cb_pa_service, s_list);
    if ((svc->caddr == addr) && (svc->rs_port == port)) 
    {
      /* HIT */
      return svc;
    }
  }

  return NULL;
}

/*
 * Create an entry into hash table for the pre-allocated real server address
 */
struct ip_vs_cb_pa_service *ip_vs_cb_pa_add_svc(__u16  protocol, __u32  caddr, __u32  rs_addr, __u16 rs_port)
{
  struct ip_vs_cb_pa_service *pa_svc = NULL;
  unsigned hash = 0;

  /*
   * Before adding new preallocated service lookup for the existing service entry
   */
  pa_svc = ip_vs_cb_pa_lookup_service(protocol, caddr, rs_port);

  if(!pa_svc)
  {
    pa_svc = (struct ip_vs_cb_pa_service*) kmalloc(sizeof(struct ip_vs_cb_pa_service), GFP_ATOMIC);

    if(pa_svc == NULL) 
    {
      IP_VS_CB_MSG("ip_vs_cb_pa_add_svc: kmalloc failed.\n");
      /*
       * No need to panic if no memory here
       * it's just an optional service
       */
    }
    else
    {
      pa_svc->caddr = caddr;
      pa_svc->rs_addr = rs_addr;
      pa_svc->rs_port = rs_port;
      atomic_set(&pa_svc->refcnt, 0);

      hash = ip_vs_cb_pa_hashkey(protocol, pa_svc->caddr, pa_svc->rs_port);

      list_add(&(pa_svc->s_list), &ip_vs_cb_pa_table[hash]);
    }
  }

  if(pa_svc)
    atomic_inc(&pa_svc->refcnt);

  return pa_svc;
}

void ip_vs_cb_pa_del_svc(struct ip_vs_cb_pa_service  *pa_svc)
{
  if(!pa_svc)
    return;

  atomic_dec(&pa_svc->refcnt);

  if(atomic_read(&pa_svc->refcnt) == 0)
  {
    /*
     * Unhash the preallocated service first
     */
    list_del(&(pa_svc->s_list));

    kfree_s(pa_svc, sizeof(struct ip_vs_cb_pa_service));
  }
}

__initfunc(int ip_vs_cb_pa_svc_init(void))
{
  int idx;                                                             

  IP_VS_CB_MSG("Initializing preallocated service module\n");

  for(idx = 0; idx < IP_VS_CB_PA_TAB_SIZE; idx++)
    INIT_LIST_HEAD(&ip_vs_cb_pa_table[idx]);                       

  return 0;
}
